# HTMLAlbum - Copyright 2000-2003 David N. Welton <davidw@dedasys.com>

# See the "LICENSE" file for terms of use and distribution.

# $Id: htmlalbum.tcl,v 1.5 2003/11/21 14:53:30 davidw Exp $

# This file contains the 'htmlalbum' package, which is actually useful
# on its own, without the GUI front-end.

package provide htmlalbum 0.2

package require html
package require xml
package require xml::c

namespace eval htmlalbum {
    set albumtitle "HTML Photo Album"

    # The directory in which the .xml file, html files, and jpegs reside.
    set albumdirectory ""
    set scriptdir [file dirname [info script]]
    # Number of images in the album.
    set imgnum 0
    set indextemplate [file join $scriptdir index.tml]
    set pagetemplate [file join $htmlalbum::scriptdir photo.tml]
}

# Clear list of files in memory.

proc htmlalbum::clearlist { } {
    variable files
    variable tree
    unset files
    unset tree
    set files {}
    array set tree {}
}

# htmlalbum::getfiles --
#
#	Start an album with the specified files.
#
# Arguments:
#	fls - a list of files to use.  A good default is [glob *.jpg].
#
# Side Effects:
#	Creates a photo album in memory, to be manipulated.
#
# Results:
#	None.

proc htmlalbum::getfiles { fls } {
    variable files $fls
    variable tree
    set tree {}
    foreach f $files {
	insert $f
    }
}

# Remove a specific number from the list of files.

proc htmlalbum::remove { number } {
    variable files
    variable tree
    set files [lreplace $files $number $number]
    set tree [lreplace $tree $number $number]
    ordertree
}

# Insert a new entry into the list of files.

proc htmlalbum::insert { flname {position end} } {
    variable tree

    set fileroot [file root $flname]
    set htmlfile "${fileroot}.html"

    array set hash [list currentfilename $htmlfile title $fileroot \
			src $flname caption $fileroot keywords ""]
    lappend tree [array get hash]
    unset hash
    ordertree
}

# Move the position of a file in the list of entries.

proc htmlalbum::move { from to } {
    variable files
    variable tree
    set fromname [lindex $files $from]
    set fromlist [lindex $tree $from]
    if { $from <= $to } {
	incr to
    } else {
	incr from
    }
    set files [lreplace [linsert $files $to $fromname] $from $from]
    set tree [lreplace [linsert $tree $to $fromlist] $from $from]
    ordertree
}

# Internal function.  After moving/ordering the list of files, run
# through it and make sure that next/previous links point to the right
# place.

proc htmlalbum::ordertree { } {
    variable files
    variable tree
    set prev ""
    set i 1
    foreach t $tree {
	array set hash $t
	if { [info exists hash(prev)] } {
	    unset hash(prev)
	}
	if { [info exists hash(next)] } {
	    unset hash(next)
	}
	array set nexthash [lindex $tree $i]
	if { [info exists nexthash(currentfilename)] } {
	    set hash(next) $nexthash(currentfilename)
	}
	if { $prev != "" } {
	    set hash(prev) $prev
	}
	set prev $hash(currentfilename)
	incr i
	lappend newtree [array get hash]

	unset hash
	unset nexthash
    }
    set tree $newtree
}

# Re-order files according to the list of numbers provided.

proc htmlalbum::order { args } {
    variable files
    variable tree
    set ordered_files ""
    set ordered_tree {}
    foreach a $args {
	lappend ordered_files [lindex $files $a]
	lappend ordered_tree [lindex $tree $a]
    }
    set files $ordered_files
    set tree $ordered_tree
    ordertree
}

# htmlalbum::gettree --
#
#	Returns a list of all the lines in the tree.
#
# Arguments:
#	None.
#
# Side Effects:
#	None.
#
# Results:
#	A list of lists.

proc htmlalbum::gettree {} {
    variable tree
    set res $tree
}

# htmlalbum::getnode --
#
#	Get the particular line out of the 'tree'.
#
# Arguments:
#	idx - index of the element to fetch.  Starts at 0.
#	attr - attribute to fetch value for.
#
# Side Effects:
#	None.
#
# Results:
#	List suitable for making a hash table with.

proc htmlalbum::getnode {idx attr} {
    variable tree
    array set hash [lindex $tree $idx]
    if { ! [info exists hash($attr)] } {
	error "htmlalbum::getnode : $attr doesn't exist, must be one of [array names hash]"
    }
    return $hash($attr)
}


# htmlalbum::getnodes --
#
#	Get particular attribute out of all nodes.
#
# Arguments:
#	attr - Attribute to fetch for all nodes.
#
# Results:
#	List of values describing the nodes.

proc htmlalbum::getnodes {attr} {
    variable tree
    set retval {}
    for {set i 0} {$i < [llength $tree]} {incr i} {
	lappend retval [htmlalbum::getnode $i $attr]
    }
    return $retval
}


# htmlalbum::gettitle --
#
#	Retrieves the album title.
#
# Results:
#	The text of the album title.

proc htmlalbum::gettitle {} {
    variable albumtitle
    return $albumtitle
}

# htmlalbum::settitle --
#
#	Set the album title
#
# Arguments:
#	title - text to use as the title.
#
# Side Effects:
#	Stores the album title.

proc htmlalbum::settitle {title} {
    variable albumtitle $title
}


# Print the current list of files, their position in the list, and
# captions.

proc htmlalbum::print { } {
    variable tree
    variable albumtitle
    variable albumdirectory
    set pos 1
    puts "Album Title: '$albumtitle', Directory: $albumdirectory (blank means current)"
    foreach t $tree {
	array set hash $t
	set caption ""
	if { [info exists hash(caption)] } {
	    set caption $hash(caption)
	}
	puts [format "Position %d, Filename %s, Caption ``%s''" $pos $hash(currentfilename) $caption]
	incr pos
	unset hash
    }
}

# Set the title for the HTML Album.

proc htmlalbum::setalbumtitle { title } {
    variable albumtitle
    set albumtitle $title
}

# htmlalbum::setnode --
#
#	Set attributes of a node.
#
# Arguments:
#	idx - Index of the node.
#	attr - Attribute to modify.
#	txt - Set the attribute to this value.
#
# Side Effects:
#	None.
#
# Results:
#	None.

proc htmlalbum::setnode { idx attr txt } {
    variable tree
    array set hash [lindex $tree $idx]
    if { ! [info exists hash($attr)] } {
	error "htmlalbum::setnode - no attribute $attr"
    }
    set hash($attr) $txt
    set tree [lreplace $tree $idx $idx [array get hash]]
}

# Set the directory in which the Album will be created.

proc htmlalbum::setdirectory { {directory ""} } {
    variable albumdirectory
    if { $directory != "" } {
	cd $directory
	set albumdirectory $directory
    }
    return $albumdirectory
}

proc htmlalbum::help { } {
    puts "HTMLAlbum Commands:
    getfiles    ?glob match?
    getxml	?xml file name?
    clearlist
    remove 	position
    insert	filename        position
    move 	position1 	position2
    order 	position1 	?position2? ...
    print
    setnode		index	attribute	text
    settitle	title
    gettitle
    setdirectory	directory
    writexml
    writehtml"
}

proc htmlalbum::? { } {
    help
}


# htmlalbum::writexml --
#
#	Writes an XML file from the in-memory tree.
#
# Arguments:
#	filename.
#
# Side Effects:
#	Creates a file, 'htmlalbum.xml' by default, in the current
#	directory.
#
# Results:
#	None.

proc htmlalbum::writexml {{filename "htmlalbum.xml"}} {
    variable files
    variable tree
    variable albumtitle
    variable albumdirectory

    if { [llength $tree] == 0 } {
	error "Need data to save - no data in memory."
    }

    set f [open "$filename" "w"]
    puts $f "<?xml version=\"1.0\"?>
    <gallery title=\"$albumtitle\" directory=\"$albumdirectory\">"
    set i 0
    foreach lst $tree {
	array set hash $lst
	set caption ""
	if { [info exists hash(caption)] } {
	    set caption $hash(caption)
	}

	puts $f "
      <page filename=\"$hash(currentfilename)\" title=\"$hash(title)\">
	<photo src=\"$hash(src)\"></photo>"

	if { [info exists hash(keywords)] } {
	    foreach key $hash(keywords) {
		puts $f "         <key>[string trim $key]</key>"
	    }
	}

	puts $f "         <caption>$caption</caption>
      </page>"

	unset hash
    }
    puts $f "</gallery>"
    close $f
}


# htmlalbum::writehtml --
#
#	Write HTML files to disk from the in-memory tree.
#
# Arguments:
#	index - name of the index file to use.  index.html is the
#	default.
#
# Side Effects:
#	Populates the directory with .html files for the selected
#	images.
#
# Results:
#	None.

proc htmlalbum::writehtml { {index "index.html"} } {
    variable files
    variable tree
    variable albumtitle
    variable pagetemplate
    #variable albumdirectory
    #if { $albumdirectory != "" } {
    #	file mkdir $albumdirectory
    #}

    # Create index.html
    set fl [open $index "w"]
    set ifl [open [setindextemplate]]
    set indextemplate [read $ifl]
    close $ifl

    set morephotos "<p align=\"center\"><b><a href=\"..\">More Photos...</a></b></p>"
    #if { $albumdirectory != "" } {
    #} else {
    #	set morephotos ""
    #}

    puts $fl [subst $indextemplate]
    close $fl

    # pull in stylesheet and png's

    catch {
	file copy -force [file join $htmlalbum::scriptdir style.css] .
    }
    catch {
	file copy [file join $htmlalbum::scriptdir right.png] .
	file copy [file join $htmlalbum::scriptdir left.png] .
	file copy [file join $htmlalbum::scriptdir up.png] .
    }

    set tfl [open [setpagetemplate]]
    set phototemplate [read $tfl]
    close $tfl
    # Create individual .html files
    foreach l $tree {
	array set hash $l
	set caption ""
	set prev ""
	set next ""
	set fl [open "$hash(currentfilename)" w]
	if { [info exists hash(prev)] } {
	    set prev " <a href=\"$hash(prev)\"><image src=\"left.png\" alt=\"Previous\"></a> "
	}
	if { [info exists hash(next)] } {
	    set next " <a href=\"$hash(next)\"><image src=\"right.png\" alt=\"Next\"></a> "
	}
	if { [info exists hash(caption)] } {
	    set caption "<p>$hash(caption)</p>"
	}

	set tmpl [subst $phototemplate]
	puts $fl $tmpl

	# copy the image jpg, as well
	#if { $albumdirectory != "" } {
	#    if { [catch { file copy -force $hash(src) $albumdirectory } err] } {
	#	error "Error: $err"
	#    }
	#}

	close $fl
	unset hash
    }
}


# htmlalbum::getxml --
#
#	Read the .xml file in to the in-memory tree representation of
#	the album.
#
# Arguments:
#	xmlfilename
#
# Side Effects:
#	Creates an album tree in memory.
#
# Results:
#	None.

proc htmlalbum::getxml { {xmlfilename "htmlalbum.xml"} } {
    variable xmlfile
    variable hashtable
    variable tree
    variable caption 0
    variable element

    set xmlfile $xmlfilename
    array set hashtable {}
    set tree {}

    set ch [open "$xmlfile"]
    set data [read $ch]
    close $ch

    array set element {caption 0 key 0}
    set parser [::xml::parser \
		    -elementstartcommand {::htmlalbum::element start} \
		    -elementendcommand {::htmlalbum::element end} \
		    -characterdatacommand ::htmlalbum::pcdata \
		    -final 1]

    $parser parse $data
}


proc htmlalbum::element {tag name {attrs {}}} {
    variable hashtable
    variable element
    variable albumtitle
    variable albumdirectory
    variable xmlfile
    variable files
    variable tree
    variable prev
    variable imgnum

    array set attrhash $attrs
    if { $tag == "start" } {
	switch -exact $name {
	    "page" {
		set hashtable(currentfilename) $attrhash(filename)
		set hashtable(title) $attrhash(title)
		set hashtable(keywords) {}
		# Ok, to set up next and prev links correctly, we need to
		# update the last entry.

		if { $imgnum > 0 } {
		    array set tmphash [lindex $tree end]
		    set tmphash(next) $attrhash(filename)
		    set tmphash(prev) $prev
		    lset tree end [array get tmphash]
		    array unset tmphash
		}
		set prev $attrhash(filename)

		incr imgnum
	    }
	    "key" {
		set element(key) 1
	    }
	    "caption" {
		set element(caption) 1
	    }
	    "photo" {
		set hashtable(src) $attrhash(src)
		lappend files $attrhash(src)
	    }
	    "gallery" {
		set albumtitle $attrhash(title)
		if { $attrhash(directory) != [file tail [pwd]] } {
		    #set albumdirectory $attrhash(directory)
		} else {
		    #set albumdirectory ""
		}
		if { $albumdirectory == "" } {
		    #set albumdirectory [file dirname $xmlfile]
		}
	    }
	}
    } elseif { $tag == "end" } {
	switch -exact $name {
	    "caption" {
		set element(caption) 0
	    }
	    "key" {
		set element(key) 0
	    }
	    "page" {
		if { ! [info exists hashtable(caption)] } {
		    set hashtable(caption) ""
		}
		lappend tree [array get hashtable]
		unset hashtable
	    }
	    "gallery" {
		# Need to finish off the last page.  Yuck...
		array set tmphash [lindex $tree end-1]
		set prev $tmphash(currentfilename)
		array unset tmphash
		array set tmphash [lindex $tree end]
		set tmphash(prev) $prev
		lset tree end [array get tmphash]
		array unset tmphash
	    }
	}
    }
    unset attrhash
}

proc htmlalbum::pcdata { pcdata } {
    variable hashtable
    variable element
    set stripped [string trim $pcdata]
    if { $element(caption) == 1 && $stripped != "" } {
	append hashtable(caption) "$stripped "
    }
    if { $element(key) == 1 } {
	lappend hashtable(keywords) $stripped
    }
    unset stripped
}

# htmlalbum::setindextemplate --
#
#	Set the template file to use for the album index.
#
# Arguments:
#	filename - if present, set the file to use.
#
# Side Effects:
#	None.
#
# Results:
#	Returns the current file to be used as the index template.

proc htmlalbum::setindextemplate { {filename ""} } {
    variable indextemplate
    if { $filename != "" } {
	set indextemplate $filename
    }
    return $indextemplate
}

# Like setindextemplate, above.

proc htmlalbum::setpagetemplate { {filename ""} } {
    variable pagetemplate
    if { $filename != "" } {
	set pagetemplate $filename
    }
    return $pagetemplate
}

namespace eval htmlalbum {
    namespace export clearlist getfiles remove insert move order print setcaption setalbumtitle settitle gettitle setdirectory help ? writexml writehtml getxml
}
